@@ -1,3 +1,32 @@
+/*  
+   ElmerGrid - A simple mesh generation and manipulation utility  
+   Copyright (C) 1995- , CSC - Scientific Computing Ltd.   
+
+   Author: Peter Rback
+   Email: Peter.Raback@csc.fi
+   Address: CSC - Scientific Computing Ltd.
+            Keilaranta 14
+            02101 Espoo, Finland
+
+   This program is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License
+   as published by the Free Software Foundation; either version 2
+   of the License, or (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+*/
+
+/* -------------------------------:  femelmer.c  :----------------------------
+   This module includes interfaces for the other Elmer programs.
+*/
+
 #include <stdio.h>
 #include <stddef.h>
 #include <stdlib.h>
@@ -18,6 +47,7 @@
 #include "femtools.h"
 #include "femtypes.h"
 #include "femknot.h"
+#include "femsolve.h"
 #include "femelmer.h"
 #include "../config.h"
 
@@ -150,6 +180,158 @@ end:
 }
 
 
+
+int FuseSolutionElmerPartitioned(char *prefix,char *outfile,int decimals,int info)
+#define MAXPAR 100
+{
+  int *noknots,*noelements,novctrs,elemcode,open;
+  int totknots,totelements,sumknots,sumelements;
+  int timesteps,i,j,k,l,step;
+  int ind[MAXNODESD1];
+  int nofiles;
+  Real r, *res, x, y, z;
+  FILE *in[MAXPAR],*out;
+  char line[MAXLINESIZE],filename[MAXFILESIZE],text[MAXNAMESIZE],outstyle[MAXFILESIZE];
+  char *cp;
+
+  for(i=0;;i++) {
+    sprintf(filename,"%s.ep.%d",prefix,i);
+    if ((in[i] = fopen(filename,"r")) == NULL) break;
+  }
+
+  nofiles = i;
+
+  if(nofiles < 2) {
+    printf("Opening of partitioned data from file %s wasn't succesfull!\n",
+	   filename);
+    return(1);
+  } else {
+    if(info) printf("Loading Elmer results from %d partitions.\n",nofiles);
+  }
+
+  noknots = Ivector(0,nofiles-1);
+  noelements = Ivector(0,nofiles-1);
+ 
+  sumknots = 0;
+  sumelements = 0;
+
+
+  for(i=0;i<nofiles;i++) {
+    fgets(line,MAXLINESIZE,in[i]);
+    if(i==0) {
+      cp = line;
+      noknots[i] = next_int(&cp);
+      noelements[i] = next_int(&cp);
+      novctrs = next_int(&cp);
+      timesteps = next_int(&cp);
+    }
+    else {
+      sscanf(line,"%d %d",&noknots[i],&noelements[i]);
+    }
+    sumknots += noknots[i];
+    sumelements += noelements[i];
+  }
+  totknots = sumknots;
+  totelements = sumelements;
+  res = Rvector(1,novctrs);
+
+  if(info) printf("There are alltogether %d nodes and %d elements.\n",totknots,sumelements);
+
+
+  AddExtension(outfile,filename,"ep");
+  if(info) printf("Saving ElmerPost data to %s.\n",filename);  
+  out = fopen(filename,"w");
+  if(out == NULL) {
+    printf("opening of file was not successful\n");
+    return(3);
+  }
+  fprintf(out,"%d %d %d %d %s %s",totknots,totelements,novctrs+1,timesteps,"scalar: Partition",cp);
+
+ 
+  if(info) printf("Reading and writing %d coordinates.\n",totknots);
+  sprintf(outstyle,"%%.%dlg %%.%dlg %%.%dlg\n",decimals,decimals,decimals);
+
+  for(j=0; j <= nofiles; j++) {
+    for(i=1; i <= noknots[j]; i++) {
+      do {
+	fgets(line,MAXLINESIZE,in[j]);
+      } while(line[0] == '#');
+
+      sscanf(line,"%le %le %le",&x,&y,&z);
+      fprintf(out,outstyle,x,y,z);
+    }
+  }
+
+  if(info) printf("Reading and writing %d element topologies.\n",totelements);
+  sumknots = 0;
+
+  for(j=0; j <= nofiles; j++) {
+    open = FALSE;
+    for(i=1; i <= noelements[j]; i++) {
+      do {
+	fgets(line,MAXLINESIZE,in[j]);
+      } while (line[0] == '#');
+
+      sscanf(line,"%s",text);
+      cp = strstr(line," ");
+
+      elemcode = next_int(&cp);
+
+      for(k=0;k< elemcode%100 ;k++) {
+	/* Dirty trick for long lines */
+	l = strspn(cp," ");
+	if( l == 0) {
+	  fgets(line,MAXLINESIZE,in[j]);
+	  cp = line;
+	}
+	ind[k] = next_int(&cp);
+      }
+      if(elemcode == 102) elemcode = 101;
+
+      fprintf(out,"%s %d",text,elemcode);
+      for(k=0;k < elemcode%100 ;k++)       
+	fprintf(out," %d",ind[k]+sumknots);
+      fprintf(out,"\n");
+    }
+    sumknots += noknots[j];
+  }
+
+  if(info) printf("Reading and writing %d degrees of freedom.\n",novctrs);
+  sprintf(outstyle,"%%.%dlg ",decimals);
+
+  for(step = 1; step <= timesteps; step++) {
+    for(k=0;k<nofiles;k++) 
+      for(i=1; i <= noknots[k]; i++) {
+	do {
+	  fgets(line,MAXLINESIZE,in[k]);
+	  if(k==0 && strstr(line,"#time")) 
+	    fprintf(out,"%s",line);
+	}
+	while (line[0] == '#');
+
+	cp = line;
+	for(j=1;j <= novctrs;j++) 
+	  res[j] = next_real(&cp);
+
+	fprintf(out,"%d ",k+1);
+	for(j=1;j <= novctrs;j++) 
+	  fprintf(out,outstyle,res[j]);
+	fprintf(out,"\n");
+      }
+  }
+
+
+  for(i=0;i<nofiles;i++) 
+    fclose(in[i]);
+  fclose(out);
+
+  if(info) printf("Successfully fused partitioned Elmer results\n");
+
+  return(0);
+}
+
+
+
 static int FindParentSide(struct FemType *data,struct BoundaryType *bound,
 			  int sideelem,int sideelemtype,int *sideind)
 {
@@ -326,8 +508,8 @@ int LoadElmerInput(struct FemType *data,struct BoundaryType *bound,
 
   sprintf(filename,"%s","mesh.nodes");
   if ((in = fopen(filename,"r")) == NULL) {
-    printf("LoadElmerInput: The opening of the nodes-file %s failed!\n",
-	   filename);
+    if(info) printf("LoadElmerInput: The opening of the nodes-file %s failed!\n",
+		    filename);
     return(2);
   }
   else 
@@ -349,7 +531,7 @@ int LoadElmerInput(struct FemType *data,struct BoundaryType *bound,
     return(3);
   }
   else 
-    printf("Loading %d Elmer elements from %s\n",noelements,filename);
+    if(info) printf("Loading %d Elmer elements from %s\n",noelements,filename);
 
   for(i=1; i <= noelements; i++) {
     fscanf(in,"%d",&dummyint);
@@ -369,8 +551,9 @@ int LoadElmerInput(struct FemType *data,struct BoundaryType *bound,
 	   filename);
     return(4);
   }
-  else 
-    printf("Loading Elmer boundaries from %s\n",filename);
+  else {
+    if(info) printf("Loading %d Elmer boundaries from %s\n",nosides,filename);
+  }
 
   AllocateBoundary(bound,nosides);
   
@@ -379,6 +562,7 @@ int LoadElmerInput(struct FemType *data,struct BoundaryType *bound,
   for(k=1; k <= nosides; k++) {
     i++;
     fscanf(in,"%d",&dummyint);
+
 #if 0
     if(k != dummyint) printf("LoadElmerInput: k=%d side=%d\n",k,dummyint);
 #endif
@@ -407,6 +591,7 @@ int LoadElmerInput(struct FemType *data,struct BoundaryType *bound,
       i--;
     }
   }
+  
   bound->nosides = i;
   fclose(in); 
 
@@ -442,8 +627,9 @@ int SaveSolutionElmer(struct FemType *data,struct BoundaryType *bound,
   
   sideelems = 0;
   if(nobound) {
-    for(i=0;i<nobound;i++)
+    for(i=0;i<nobound;i++) {
       if(bound[i].created) sideelems += bound[i].nosides; 
+    }
   }
 
   noknots = data->noknots;
@@ -665,21 +851,26 @@ int SaveElmerInput(struct FemType *data,
 
     fprintf(out,"%d %d %d",i,data->material[i],elemtype);
 
-    if(data->pelems == 1) {
-      if(data->pelemtypes[i] > 0) fprintf(out,"p%d",data->pelemtypes[i]);
-    }
-    else if(data->pelems == 2) {
+    if(data->pelems) {
       j = data->pelemtypes[i];
+
       k = j; j=j/10; k=k-10*j;
       if(k!=1) fprintf(out,"n%d",k);
+
       k = j; j=j/10; k=k-10*j;
       if(k!=0) fprintf(out,"e%d",k);
+
       k = j; j=j/10; k=k-10*j;
       if(k!=0) fprintf(out,"f%d",k);
+
       k = j; j=j/10; k=k-10*j;
       if(k!=0) fprintf(out,"d%d",k);
+
       k = j; j=j/100; k=k-100*j;
       if(k!=0) fprintf(out,"b%d",k);
+
+      k = j; j=j/100; k=k-100*j;
+      if(k!=0) fprintf(out,"p%d",k);
     }
       
     bulktypes[elemtype] += 1;
@@ -1848,9 +2039,10 @@ int OptimizePartitioning(struct FemType *data,struct BoundaryType *bound,
 
 #if 1
   for(i=1;i<=noknots;i++) {
+
     ind = i;
     if(periodic) ind = indxper[ind];
-    
+
     if(neededtimes[ind] > 1) {
       j = neededtable[ind][1];
       k = neededtable[ind][2];
@@ -1866,6 +2058,15 @@ int OptimizePartitioning(struct FemType *data,struct BoundaryType *bound,
       }
     }
   }
+
+  /* A posteriori correction, don't know if this just corrects the symptom */
+  if(periodic) {
+    for(i=1;i<=noknots;i++) {
+      ind = indxper[i];
+      if(i != ind) neededby[i] = neededby[ind]; 
+    }
+  }
+
 #else 
   for(i=1;i<=noknots;i++)
     neededby[i] = nodepart[i];
@@ -1961,8 +2162,12 @@ optimizeownership:
   
   /* Check the neededvector to be on the safe side */
   for(i=1;i<=noknots;i++) {
+    if(periodic) {
+      ind = indxper[i];
+      if(ind != i) continue;
+    }
     k = 0;
-    for(j=1;j<=neededtimes[i];j++)
+    for(j=1;j<=neededtimes[i];j++) 
       if(neededtable[i][j] == neededby[i]) k++;
     if(k!=1) printf("Node %d is owned %d times by partition %d\n",i,k,neededby[i]);
   }
