@@ -1,9 +1,37 @@
+/*  
+   ElmerGrid - A simple mesh generation and manipulation utility  
+   Copyright (C) 1995- , CSC - Scientific Computing Ltd.   
+
+   Author: Peter Rback
+   Email: Peter.Raback@csc.fi
+   Address: CSC - Scientific Computing Ltd.
+            Keilaranta 14
+            02101 Espoo, Finland
+
+   This program is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License
+   as published by the Free Software Foundation; either version 2
+   of the License, or (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+*/
+
+/* --------------------:  femfile.c  :-------------------------- */
+
 #include <stdio.h>
 #include <math.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <ctype.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "nrutil.h"
 #include "common.h"
@@ -49,6 +77,28 @@ static int Getrow(char *line1,FILE *io,int upper)
 }
 
 
+static int Comsolrow(char *line1,FILE *io) 
+{
+  int i,isend;
+  char line0[MAXLINESIZE],*charend;
+
+  for(i=0;i<MAXLINESIZE;i++) 
+    line0[i] = ' ';
+
+ newline:
+
+  charend = fgets(line0,MAXLINESIZE,io);
+  isend = (charend == NULL);
+
+  if(isend) return(1);
+
+  for(i=0;i<MAXLINESIZE;i++) line1[i] = line0[i];    
+
+  return(0);
+}
+
+
+
 static void FindPointParents(struct FemType *data,struct BoundaryType *bound,
 			    int boundarynodes,int *nodeindx,int *boundindx,int info)
 {
@@ -82,11 +132,11 @@ static void FindPointParents(struct FemType *data,struct BoundaryType *bound,
   for(elemind=1;elemind<=data->noelements;elemind++) {
     elemtype = data->elementtypes[elemind];
     elemsides = elemtype % 100;
-    for(i=0;i<elemsides;i++) 
+    for(i=0;i<elemsides;i++) {
       elemhits[data->topology[elemind][i]] += 1;
+    }
   }
-    
- 
+
 
   for(boundarytype=minboundary;boundarytype <= maxboundary;boundarytype++) {
     int boundfirst,bchits,bcsame,sideelemtype2;
@@ -102,6 +152,7 @@ static void FindPointParents(struct FemType *data,struct BoundaryType *bound,
 	indx[nodeindx[i]] = TRUE; 
     }
 
+
     for(elemind=1;elemind<=data->noelements;elemind++) {
       elemtype = data->elementtypes[elemind];
       elemsides = elemtype / 100;
@@ -379,7 +430,7 @@ omstart:
 	
 	if(maxnodes < elemcode%100) maxnodes = elemcode%100;
 	mode = 3;
-	if(0) printf("Loading elements of type %d starting from element %d.\n",
+	if(1) printf("Loading elements of type %d starting from element %d.\n",
 			elemcode,noelements);
       }
       else if(strstr(line,"BOUNDARY")) {
@@ -1024,6 +1075,7 @@ int LoadFidapInput(struct FemType *data,char *prefix,int info)
   elemcode = 0;
   maxnodes = 4;
   totelems = 0;
+  maxentity = 0;
 
   for(;;) {
     isio = getline;
@@ -1049,7 +1101,7 @@ int LoadFidapInput(struct FemType *data,char *prefix,int info)
     switch (mode) {
 
     case 1: 
-      if(info) printf("Loading FIDAP input file:\n");
+      if(info) printf("Loading FIDAP input file %s\n",filename);
       getline;
       if(info) printf("Name of the case: %s",line);
       mode = 0;
@@ -1057,7 +1109,7 @@ int LoadFidapInput(struct FemType *data,char *prefix,int info)
 
     case 2:
       getline;   
-      if(info) printf("reading the header info\n");
+      if(0) printf("reading the header info\n");
       sscanf(line,"%d%d%d%d%d",&noknots,&noelements,
 	     &nogroups,&dim,&novel);
       data->noknots = noknots;
@@ -1129,7 +1181,7 @@ int LoadFidapInput(struct FemType *data,char *prefix,int info)
 	  data->topology = topology;
 	}
 
-	if(info) printf("reading %d element topologies with %d nodes for %s\n",
+	if(0) printf("reading %d element topologies with %d nodes for %s\n",
 			elems,nodes,entityname);
 
 	for(entity=1;entity<=maxentity;entity++) {
@@ -1140,6 +1192,7 @@ int LoadFidapInput(struct FemType *data,char *prefix,int info)
 	if(entity > maxentity) {
 	  maxentity++;
 	  strcpy(entitylist[entity],entityname);
+	  if(info) printf("Found new entity: %s\n",entityname);
 	}
 
 	for(i=totelems+1;i<=totelems+elems;i++) {
@@ -1225,7 +1278,11 @@ end:
   }
 
   fclose(in);
+  
+  if(info) printf("Finished reading the Fidap neutral file\n");
+
   return(0);
+
 }
 
 
@@ -1343,9 +1400,10 @@ int LoadAnsysInput(struct FemType *data,struct BoundaryType *bound,
 {
   int noknots,noelements,nosides,elemcode,sidetype,currenttype;
   int elementtypes,sideind[MAXNODESD1],tottypes,elementtype;
-  int maxindx,*indx,*revindx,topology[100];
+  int maxindx,*indx,*revindx,topology[100],ind;
   int i,j,k,l,imax,grp,dummyint,*nodeindx,*boundindx,boundarynodes,maxnodes;
   int noansystypes,*ansysdim,*ansysnodes,*ansystypes,boundarytypes;
+  int debug;
   Real x,y,z,r;
   FILE *in;
   char *cp,line[MAXLINESIZE],filename[MAXFILESIZE],
@@ -1542,12 +1600,16 @@ int LoadAnsysInput(struct FemType *data,struct BoundaryType *bound,
     printf("Loading %d Ansys elements from %s\n",noelements,filename);
 
   for(j=1;j<=noelements;j++) {
+
     getline; cp=line;
-    for(i=0;i<8;i++)
-      topology[i] = revindx[next_int(&cp)];
+
+    for(i=0;i<8;i++) {
+      ind = next_int(&cp);
+      if(cp[0] == '.') cp++;
+      topology[i] = revindx[ind];
+    }
 
     data->material[j] = next_int(&cp);
-
     currenttype = next_int(&cp);
     
     for(k=1;k<=noansystypes;k++) 
@@ -1562,8 +1624,13 @@ int LoadAnsysInput(struct FemType *data,struct BoundaryType *bound,
       else 
 	imax = 20;
 
-      for(i=8;i<imax;i++) 
-	topology[i] = revindx[next_int(&cp)];
+      if(debug) printf("imax = %d\n",imax);
+
+      for(i=8;i<imax;i++) {
+	ind = next_int(&cp);
+	if(cp[0] == '.') cp++;
+	topology[i] = revindx[ind];
+      }
     }
 
 #if 0
@@ -1941,113 +2008,6 @@ int LoadFemlabMesh(struct FemType *data,struct BoundaryType *bound,
 
 
 
-int LoadFemlab3Mesh(struct FemType *data,struct BoundaryType *bound,
-		   char *prefix,int info)
-     /* This procedure reads the FEM mesh as written by Femlab3 */
-{
-  int noknots,noelements,nosides,elemcode,sideelem,sidetype,currenttype;
-  int elementtypes,sideind[MAXNODESD1],tottypes,elementtype;
-  int i,j,k,l,imax,grp,dummyint,**nodeindx,*boundindx,boundarynodes,maxnodes;
-  int dim,nonodes,sidenodes,mode;
-  Real x,y,z,dummy;
-  FILE *in;
-  char *cp,line[MAXLINESIZE],filename[MAXFILESIZE],
-    text[MAXNAMESIZE],directoryname[MAXFILESIZE];
-
-
-  sprintf(filename,"%s",prefix);
-  if ((in = fopen(filename,"r")) == NULL) {
-    sprintf(filename,"%s.p",prefix);
-    if ((in = fopen(filename,"r")) == NULL) {
-      printf("LoadFemlabInput: The opening of the header-file %s failed!\n",
-	     filename);
-      return(1);
-    }
-  }
-
-  mode = 0;
-  for(;;) {
-    getline;
-    
-    if(strstr(line,"Coordinates")) {
-      mode = 1;
-      noknots = 0;
-      getline;
-    }
-
-    if(strstr(line,"Elements")) {
-      mode = 2;
-      noelements= 0;
-      getline;
-    }
-    if(strstr(line,"Data")) {
-      mode = 3;
-    }
-
-    if(mode == 1) noknots = noknots + 1;
-    if(mode == 2) noelements = noelements + 1;
-    if(mode == 3) break;
-  }
-  fclose(in);
-
-  printf("Femlab file has %d elements and %d nodes.\n",noelements,noknots);
-
-  in = fopen(filename,"r");
-
-
-  InitializeKnots(data);
-
-  /* Even 2D elements may form a 3D object! */
-  data->dim = 2;
-  data->maxnodes = 3;
-  data->noknots = noknots;
-  data->noelements = noelements;
-  AllocateKnots(data);
-
-  mode = 0;
-  for(;;) {
-    getline;
-    
-    if(strstr(line,"Coordinates")) {
-      mode = 1;
-      noknots = 0;
-      getline;
-    }
-
-    if(strstr(line,"Elements")) {
-      mode = 2;
-      noelements = 0;
-      getline;
-    }
-    if(strstr(line,"Data")) {
-      mode = 3;
-    }
-
-    if(mode == 1) {
-      noknots = noknots + 1;
-      cp=line;
-      x = next_real(&cp);
-      y = next_real(&cp);
-      data->x[noknots] = x;
-      data->y[noknots] = y;      
-    }
-    if(mode == 2) {
-      cp = line;
-      noelements = noelements + 1;
-      for(i=0;i<3;i++) {
-	data->topology[noelements][i] = next_int(&cp);
-      }      
-      data->elementtypes[noelements] = 303;
-      data->material[noelements] = 1;      
-    }
-    if(mode == 3) break;
-  }
-  fclose(in);
-
-  return(0);
-}
-
-
 
 static void ReorderFieldviewNodes(struct FemType *data,int *oldtopology,
 				  int element,int dim,int nodes) 
@@ -2860,6 +2820,214 @@ end:
 }
 
 
+int LoadComsolMesh(struct FemType *data,char *prefix,int info)
+/* Load the grid in Comsol Multiphysics mesh format */
+{
+  int noknots,noelements,elemcode,maxnodes,material,foundsame;
+  int mode,allocated,nvalue,maxknot,nosides,sideelemtype;
+  int boundarytype,materialtype,boundarynodes,side,parent,elemsides;
+  int dim, elemnodes, elembasis, elemtype, bulkdone, usedmax,hits;
+  int label,debug,offset,domains,mindom,minbc,elemdim;
+  char filename[MAXFILESIZE],line[MAXLINESIZE],*cp;
+  int i,j,k,l,n,ind,inds[MAXNODESD2],sideind[MAXNODESD1];
+  FILE *in;
+  Real x,y,z;
+
+
+  strcpy(filename,prefix);
+  if ((in = fopen(filename,"r")) == NULL) {
+    AddExtension(prefix,filename,"mphtxt");
+    if ((in = fopen(filename,"r")) == NULL) {
+      printf("LoadComsolMesh: opening of the Comsol mesh file '%s' wasn't succesfull !\n",
+	     filename);
+      return(1);
+    }
+  }
+
+  printf("Reading mesh from Comsol mesh file %s.\n",filename);
+  InitializeKnots(data);
+
+  debug = FALSE;
+  allocated = FALSE;
+
+  mindom = 1000;
+  minbc = 1000;
+  offset = 1;
+
+omstart:
+
+  maxnodes = 0;
+  noknots = 0;
+  noelements = 0;
+  elemcode = 0;
+  material = 0;
+  domains = 0;
+
+
+  for(;;) {
+
+    if(Comsolrow(line,in)) goto end;
+    if(!line) goto end;
+
+    if(strstr(line,"# sdim")) {
+      cp = line;
+      dim = next_int(&cp);
+      if(debug) printf("dim=%d\n",dim);
+    }
+
+    else if(strstr(line,"# number of mesh points")) {
+      cp = line;
+      noknots = next_int(&cp);
+      if(debug) printf("noknots=%d\n",noknots);
+    }
+
+    else if(strstr(line,"# lowest mesh point index")) {
+      cp = line;
+      offset = 1 - next_int(&cp);
+      if(debug) printf("offset=%d\n",offset);
+    }
+
+    else if(strstr(line,"# type name")) {
+      if(strstr(line,"vtx")) elembasis = 100;
+      if(strstr(line,"edg")) elembasis = 200;
+      if(strstr(line,"tri")) elembasis = 300;
+      if(strstr(line,"tet")) elembasis = 500;
+      if(debug) printf("elembasis=%d\n",elembasis);     
+    }
+
+    else if(strstr(line,"# number of nodes per element")) {
+      cp = line;
+      elemnodes = next_int(&cp);
+      if(elemnodes > maxnodes) maxnodes = elemnodes;      
+      if(debug) printf("elemnodes=%d\n",elemnodes);           
+    }
+
+    else if(strstr(line,"# Mesh point coordinates")) {
+      printf("Loading %d coordinates\n",noknots);
+
+      for(i=1;i<=noknots;i++) {
+	Comsolrow(line,in);	
+
+	if(allocated) {
+	  cp = line;
+	  data->x[i] = next_real(&cp);
+	  data->y[i] = next_real(&cp);
+	  if(dim == 3) data->z[i] = next_real(&cp);
+	}
+      }
+    }
+
+    else if(strstr(line,"# number of elements")) {
+
+      cp = line;
+      k = next_int(&cp);
+
+      Comsolrow(line,in);	            
+      elemtype = elemnodes + elembasis;
+
+      if(elembasis > 400) 
+	elemdim = 3;
+      else if(elembasis > 200)
+	elemdim = 2;
+      else 
+	elemdim = 1;
+
+      if(debug) printf("Loading %d elements of type %d\n",k,elemtype);
+      domains = noelements;
+
+      for(i=1;i<=k;i++) {
+	Comsolrow(line,in);	
+
+	if(dim == 3 && elembasis < 300) continue;
+	if(dim == 2 && elembasis < 200) continue;
+
+	noelements = noelements + 1;
+	if(allocated) {
+	  cp = line;
+	  for(j=0;j<elemnodes;j++)
+	    data->topology[noelements][j] = next_int(&cp) + offset;
+	  data->elementtypes[noelements] = elemtype;
+	  data->material[noelements] = 1;	  
+	}
+      }
+    }
+
+    else if(strstr(line,"# number of domains")) {
+
+      cp = line;
+      k = next_int(&cp);
+
+      Comsolrow(line,in);	            
+      if(debug) printf("Loading %d domains for the elements\n",k);
+
+      for(i=1;i<=k;i++) {
+	Comsolrow(line,in);	
+
+	if(dim == 3 && elembasis < 300) continue;
+	if(dim == 2 && elembasis < 200) continue;
+
+	domains = domains + 1;
+	cp = line;
+	material = next_int(&cp);
+
+	if(allocated) {
+	  if(elemdim < dim) 
+	    material = material - minbc + 1;
+	  else 
+	    material = material - mindom + 1;
+	  data->material[domains] = material;	  
+	}
+	else {
+	  if(elemdim < dim) {
+	    if(minbc > material) minbc = material;
+	  }
+	  else {
+	    if(mindom > material) mindom = material;	  
+	  }	  
+	}
+
+      }
+    }
+
+    else if(strstr(line,"#")) {
+      if(debug) printf("Unused command:  %s",line);
+    }
+
+  }
+
+end:
+
+  if(!allocated) {
+
+    if(noknots == 0 || noelements == 0 || maxnodes == 0) {
+       printf("Invalid mesh consits of %d knots and %d %d-node elements.\n",
+	     noknots,noelements,maxnodes);     
+       close(in);
+       return(2);
+    }
+
+    rewind(in);
+    data->noknots = noknots;
+    data->noelements = noelements;
+    data->maxnodes = maxnodes;
+    data->dim = dim;
+    
+    if(info) {
+      printf("Allocating for %d knots and %d %d-node elements.\n",
+	     noknots,noelements,maxnodes);
+    }  
+    AllocateKnots(data);
+    allocated = TRUE;
+
+    goto omstart;    
+  }
+  close(in);
+
+  if(info) printf("The Comsol mesh was loaded from file %s.\n\n",filename);
+  return(0);
+}
+
+
 
 static int GmshToElmerType(int gmshtype)
 {
