@@ -1,3 +1,28 @@
+/*  
+   ElmerGrid - A simple mesh generation and manipulation utility  
+   Copyright (C) 1995- , CSC - Scientific Computing Ltd.   
+
+   Author: Peter Rback
+   Email: Peter.Raback@csc.fi
+   Address: CSC - Scientific Computing Ltd.
+            Keilaranta 14
+            02101 Espoo, Finland
+
+   This program is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License
+   as published by the Free Software Foundation; either version 2
+   of the License, or (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+*/
+
 /****************************************************************************
 *                                                                           *  
 *                              Elmergrid                                    *
@@ -12,17 +37,9 @@
 *  been designed and with new features the code has eventually become very  *
 *  dirty.                                                                   *
 *                                                                           *
-*  (c) Peter Rback, CSC - Scientific Computing, 1996-2005                  *
-*                                                                           *
-*      Email: Peter.Raback@csc.fi                                           *
-*      Fax:   +358-9-457 2302                                               *
-*      Phone: +358-9-457 2080                                               *
-*                                                                           *
 ****************************************************************************/
 
 
-
-
 #include <stdio.h>
 #include <stddef.h>
 #include <stdlib.h>
@@ -68,12 +85,13 @@ static void Instructions()
   printf("6)  .fil      : Abaqus output format\n");
   printf("7)  .FDNEUT   : Fidap neutral file\n");
   printf("8)  .d        : Easymesh input format\n");
-  printf("9)  .femlab   : Femlab mesh format\n");
+  printf("9)  .mphtxt   : Comsol Multiphysics (was Femlab) mesh format\n");
   printf("10) .dat      : Fieldview format\n");
   printf("11) .node,.ele: Triangle 2D mesh format\n");
   printf("12) .mesh     : Medit mesh format\n");
   printf("13) .msh      : GID mesh format\n");
   printf("14) .msh      : Gmsh mesh format\n");
+  printf("15) .ep.i     : Partitioned ElmerPost format\n");
 
   printf("\nThe second parameter defines the output file format:\n");
   printf("1)  .grd      : ElmerGrid file format\n");
@@ -113,9 +131,11 @@ static void Instructions()
   printf("-mirror int[3]       : copy the mesh around the origin in coordinate directions\n");
   printf("-unite               : the meshes will be united\n");
   printf("-polar real          : map 2D mesh to a cylindrical shell with given radius\n");
+  printf("-cylinder            : map 2D/3D cylindrical mesh to a cartesian mesh\n");
   printf("-reduce int[2]       : reduce element order at material interval [int1 int2]\n");
   printf("-increase            : increase element order from one to two\n");
-  printf("-pelem int[3]        : make p-elements of power int3 at interval [int1 int2]\n");
+  printf("-pelem int[3]        : p-elements of power int3 at interval [int1 int2]\n");
+  printf("-belem int[3]        : set bubble dofs to int3 at interval [int1 int2]\n");
   printf("-partition int[3]    : the mesh will be partitioned in main directions\n");
   printf("-partorder real[3]   : in the above method, the direction of the ordering\n");
 #if PARTMETIS
@@ -129,7 +149,7 @@ static void Instructions()
   printf("-bulktype int[3]     : set material types in interval [int1 int2] to type int3\n");
   printf("-boundtype int[3]    : set sidetypes in interval [int1 int2] to type int3\n");
   printf("-layer int[2] real[2]: make a boundary layer for given boundary\n");
-  printf("-layermove           : in the previous case move the layer to the original boundary\n");
+  printf("-layermove int       : apply Jacobi filter int times to move the layered mesh\n");
   printf("-3d / -2d / -1d      : mesh is 3, 2 or 1-dimensional\n");
   printf("-isoparam            : ensure that higher order elements are convex\n");
   printf("-nobound             : disable saving of boundary elements\n");
@@ -158,6 +178,7 @@ void InitParameters(struct ElmergridType *eg)
   eg->triangles = FALSE;
   eg->rotate = FALSE;
   eg->polar = FALSE;
+  eg->cylinder = FALSE;
   eg->layers = 0;
   eg->layereps = 0.0;
   eg->layermove = 0;
@@ -177,6 +198,7 @@ void InitParameters(struct ElmergridType *eg)
   eg->partorder = FALSE;
   eg->findsides = FALSE;
   eg->pelems = 0;
+  eg->belems = 0;
   eg->saveboundaries = TRUE;
   eg->merge = FALSE;
   eg->bcoffset = FALSE;
@@ -216,9 +238,13 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
   else if(strcmp(argv[1],"Fidap") == 0) eg->inmethod = 7;
   else if(strcmp(argv[1],"Easymesh") == 0) eg->inmethod = 8;
   else if(strcmp(argv[1],"Femlab") == 0) eg->inmethod = 9;
+  else if(strcmp(argv[1],"Comsol") == 0) eg->inmethod = 9;
   else if(strcmp(argv[1],"Fieldview") == 0) eg->inmethod = 10;
   else if(strcmp(argv[1],"Triangle") == 0) eg->inmethod = 11;
   else if(strcmp(argv[1],"Medit") == 0) eg->inmethod = 12;
+  else if(strcmp(argv[1],"Gid") == 0) eg->inmethod = 13;
+  else if(strcmp(argv[1],"Gmsh") == 0) eg->inmethod = 14;
+  else if(strcmp(argv[1],"Partitioned") == 0) eg->inmethod = 15;
   else eg->inmethod = atoi(argv[1]);
 
   if(strcmp(argv[2],"ElmerGrid") == 0) eg->outmethod = 1;
@@ -251,7 +277,6 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       }
       else {
 	strcpy(eg->filesout[0],argv[arg+1]);
-	printf("Output will be saved to file %s.\n",eg->filesout[0]);
       }
     }
 
@@ -390,6 +415,11 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       }
     }
 
+    if(strcmp(argv[arg],"-cylinder") == 0) {
+      eg->cylinder = TRUE;
+      printf("Making transformation from cylindrical to cartesian coordinates.\n");
+    }
+
     if(strcmp(argv[arg],"-reduce") == 0) {
       if(arg+2 >= argc) {
 	printf("Give two material for the interval.\n");
@@ -409,6 +439,11 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
 	eg->pelemmap[3*eg->pelems+i-1-arg] = atoi(argv[i]);
       eg->pelems++;
     } 
+    if(strcmp(argv[arg],"-belem") == 0) {
+      for(i=arg+1;i<argc && strncmp(argv[i],"-",1); i++) 
+	eg->belemmap[3*eg->belems+i-1-arg] = atoi(argv[i]);
+      eg->belems++;
+    } 
     if(strcmp(argv[arg],"-partition") == 0) {
       if(arg+dim >= argc) {
 	printf("The number of partitions in %d dims is required as paramaters.\n",dim);
@@ -520,7 +555,13 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       }
     }
     if(strcmp(argv[arg],"-layermove") == 0) {
-      eg->layermove = TRUE;
+      if(arg+1 >= argc) {
+	printf("Give maximum number of Jacobi filters.\n");
+ 	return(11);
+      }
+      else {
+	eg->layermove = atoi(argv[arg+1]);
+      }
     }
 
     if(strcmp(argv[arg],"-3d") == 0) {
@@ -624,6 +665,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
 
 
   for(;;) {
+
     if(GetCommand(command,params,in)) {
       printf("Reached the end of command file\n");
       goto end;
@@ -640,28 +682,30 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
 	for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
 
 	if(strstr(params,"ELMERGRID")) eg->inmethod = 1;
-	if(strstr(params,"ELMERSOLVER")) eg->inmethod = 2;
-	if(strstr(params,"ELMERPOST")) eg->inmethod = 3;
-	if(strstr(params,"ANSYS")) eg->inmethod = 4;
-	if(strstr(params,"IDEAS")) eg->inmethod = 5;
-	if(strstr(params,"ABAQUS")) eg->inmethod = 6;
-	if(strstr(params,"FIDAP")) eg->inmethod = 7;
-	if(strstr(params,"EASYMESH")) eg->inmethod = 8;
-	if(strstr(params,"FEMLAB")) eg->inmethod = 9;
-	if(strstr(params,"FIELDVIEW")) eg->inmethod = 10;
-	if(strstr(params,"TRIANGLE")) eg->inmethod = 11;
-	if(strstr(params,"MEDIT")) eg->inmethod = 12;
-	if(strstr(params,"GID")) eg->inmethod = 13;
-	if(strstr(params,"GMSH")) eg->inmethod = 14;
+	else if(strstr(params,"ELMERSOLVER")) eg->inmethod = 2;
+	else if(strstr(params,"ELMERPOST")) eg->inmethod = 3;
+	else if(strstr(params,"ANSYS")) eg->inmethod = 4;
+	else if(strstr(params,"IDEAS")) eg->inmethod = 5;
+	else if(strstr(params,"ABAQUS")) eg->inmethod = 6;
+	else if(strstr(params,"FIDAP")) eg->inmethod = 7;
+	else if(strstr(params,"EASYMESH")) eg->inmethod = 8;
+	else if(strstr(params,"FEMLAB")) eg->inmethod = 9;
+	else if(strstr(params,"COMSOL")) eg->inmethod = 9;
+	else if(strstr(params,"FIELDVIEW")) eg->inmethod = 10;
+	else if(strstr(params,"TRIANGLE")) eg->inmethod = 11;
+	else if(strstr(params,"MEDIT")) eg->inmethod = 12;
+	else if(strstr(params,"GID")) eg->inmethod = 13;
+	else if(strstr(params,"GMSH")) eg->inmethod = 14;
+	else if(strstr(params,"PARTITIONED")) eg->inmethod = 15;
 	else sscanf(params,"%d",&eg->inmethod);
       }
       else if(strstr(command,"OUTPUT MODE")) {
 	for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
 
 	if(strstr(params,"ELMERGRID")) eg->outmethod = 1;
-	if(strstr(params,"ELMERSOLVER")) eg->outmethod = 2;
-	if(strstr(params,"ELMERPOST")) eg->outmethod = 3;
-	if(strstr(params,"FASTCAP")) eg->outmethod = 13;
+	else if(strstr(params,"ELMERSOLVER")) eg->outmethod = 2;
+	else if(strstr(params,"ELMERPOST")) eg->outmethod = 3;
+	else if(strstr(params,"FASTCAP")) eg->outmethod = 13;
 	else sscanf(params,"%d",&eg->outmethod);	
       }
     }    
@@ -710,12 +754,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
     }
     else if(strstr(command,"ROTATE MESH")) {
       eg->rotate = TRUE;
-      if(eg->dim == 1) 
-	sscanf(params,"%le",&eg->crotate[0]);
-      else if(eg->dim == 2) 
-	sscanf(params,"%le%le",&eg->crotate[0],&eg->crotate[1]);
-      else if(eg->dim == 3) 
-	sscanf(params,"%le%le%le",&eg->crotate[0],&eg->crotate[1],&eg->crotate[2]);
+      sscanf(params,"%le%le%le",&eg->crotate[0],&eg->crotate[1],&eg->crotate[2]);
     }
     else if(strstr(command,"CLONE")) {
       if(strstr(command,"CLONE SIZE")) {
@@ -747,6 +786,10 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       eg->polar = TRUE;
       sscanf(params,"%le",&eg->polarradius);
     }
+    else if(strstr(command,"CYLINDER")) {
+      for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
+      if(strstr(params,"TRUE")) eg->cylinder = TRUE;      
+    }
     else if(strstr(command,"REDUCE DEGREE")) {
       eg->reduce = TRUE;
       sscanf(params,"%d%d",&eg->reducemat1,&eg->reducemat2);
@@ -763,12 +806,13 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
 	for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
 	if(strstr(params,"END")) break;
 		
-	sscanf(params,"%d%d%d%d%d%d",
-	       &eg->advancedelem[6*i],&eg->advancedelem[6*i+1],&eg->advancedelem[6*i+2],
-	       &eg->advancedelem[6*i+3],&eg->advancedelem[6*i+4],&eg->advancedelem[6*i+5]);
+	sscanf(params,"%d%d%d%d%d%d%d",
+	       &eg->advancedelem[7*i],&eg->advancedelem[7*i+1],&eg->advancedelem[7*i+2],
+	       &eg->advancedelem[7*i+3],&eg->advancedelem[7*i+4],&eg->advancedelem[7*i+5],
+	       &eg->advancedelem[7*i+6]);
       }  
       eg->advancedmat = i;
-      printf("Found %d definitions for p-elements.\n",i);
+      printf("Found %d definitions for advanced elements.\n",i);
     }
     else if(strstr(command,"POWER ELEMENTS")) {
       printf("Loading p-type element definitions\n");
@@ -783,6 +827,19 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       eg->pelems = i;
       printf("Found %d definitions for p-elements.\n",i);
     }
+    else if(strstr(command,"BUBBLE ELEMENTS")) {
+      printf("Loading bubble element definitions\n");
+      
+      for(i=0;i<MAXMATERIALS;i++) {
+	if(i>0) Getline(params,in);
+	for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
+	if(strstr(params,"END")) break;
+	sscanf(params,"%d%d%d",
+	       &eg->belemmap[3*i],&eg->belemmap[3*i+1],&eg->belemmap[3*i+2]);
+      }  
+      eg->belems = i;
+      printf("Found %d definitions for bubble elements.\n",i);
+    }
     else if(strstr(command,"METIS")) {
 #if PARTMETIS
       sscanf(params,"%d",&eg->metis);
@@ -857,8 +914,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
 
     else if(strstr(command,"BOUNDARY LAYER")) {
       if(strstr(command,"BOUNDARY LAYER MOVE")) {
-	for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
-	if(strstr(params,"TRUE")) eg->layermove = TRUE;
+	sscanf(params,"%d",&eg->layermove);
       }
       else if(strstr(command,"BOUNDARY LAYER EPSILON")) {
 	sscanf(params,"%le",&eg->layereps);
@@ -893,7 +949,6 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
     }
     else if(strstr(command,"EXTRUDED")) {
       grid->dimension = 3;
-      grid->autoratio = FALSE;    
 
       if(strstr(command,"EXTRUDED DIVISIONS")) {
 	sscanf(params,"%d",&grid->zcells);		
@@ -905,6 +960,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       else if(strstr(command,"EXTRUDED ELEMENTS")) {
 	cp = params;
 	for(i=1;i<=grid->zcells;i++) grid->zelems[i] = next_int(&cp);
+	grid->autoratio = FALSE;    
       }
       else if(strstr(command,"EXTRUDED RATIOS")) {
 	cp = params;
@@ -1077,17 +1133,22 @@ int main(int argc, char *argv[])
       printf("Because file %s didn't exist, it was created for you.\n",eg.filesin[nofile]);
       Goodbye();
     }
-
     LoadCommands(eg.filesin[nofile],&eg,grids,1,info); 
     break;
 
   case 2: 
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if(LoadElmerInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],info))
       Goodbye();
     nomeshes = nofile+1;
-    if(boundaries[nofile]->created) data[nofile].noboundaries = 1;
+    if(0 && boundaries[nofile]->created) {
+      data[nofile].noboundaries = 1;
+    }
     break;
 
   case 3: 
@@ -1107,6 +1168,10 @@ int main(int argc, char *argv[])
   case 5: 
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if(LoadAbaqusInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],TRUE)) 
       Goodbye();
     nomeshes = nofile+1;
@@ -1123,7 +1188,10 @@ int main(int argc, char *argv[])
       Goodbye();
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     ElementsToBoundaryConditions(&(data[nofile]),boundaries[nofile],TRUE);
 
     nomeshes = nofile+1;
@@ -1144,6 +1212,10 @@ int main(int argc, char *argv[])
       AllocateKnots(&(data[nofile]));
       boundaries[nofile] = (struct BoundaryType*)
 	malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
+      for(i=0;i<MAXBOUNDARIES;i++) {
+	boundaries[nofile][i].created = FALSE; 
+	boundaries[nofile][i].nosides = 0;
+      }
 
       if(EasymeshCopy(&(data[nofile]),boundaries[nofile]))
 	Goodbye();
@@ -1155,25 +1227,39 @@ int main(int argc, char *argv[])
  case 9:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    if(LoadFemlabMesh(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],info)) 
-      Goodbye();
-    nomeshes = nofile + 1;
-    break;
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
+   
+    if(LoadComsolMesh(&(data[nofile]),eg.filesin[nofile],info)) {
 
- case 99:
-    boundaries[nofile] = (struct BoundaryType*)
-      malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    if(LoadFemlab3Mesh(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],info)) 
-      Goodbye();
-    nomeshes = nofile + 1;
-    break;
+      printf("\n***********************************************************************************\n");
+      printf("The reading of Comsol mesh file seems to have failed\n");
+      printf("Trying out a previous version that requires the use of savemesh.m utility in Matlab\n");
+      printf("The recommended way to export meshes from Femlab to Elmer is the .mphtxt format\n"); 
+      printf("***********************************************************************************\n\n");
 
+      if(LoadFemlabMesh(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],info)) 
+	Goodbye();
+      nomeshes = nofile + 1;      
+    }
+
+    ElementsToBoundaryConditions(&(data[nofile]),boundaries[nofile],TRUE);
+    nomeshes = nofile + 1;
+
+    break;
 
   case 10:
     if(LoadFieldviewInput(&(data[nofile]),eg.filesin[nofile],TRUE))
       Goodbye();
     boundaries[nofile] = (struct BoundaryType*)
-      malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
+      malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	    
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
+
     ElementsToBoundaryConditions(&(data[nofile]),boundaries[nofile],TRUE);
     nomeshes = 1;
     break;
@@ -1181,8 +1267,10 @@ int main(int argc, char *argv[])
   case 11:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    for(j=0;j<MAXBOUNDARIES;j++) 
-      boundaries[nofile][j].created = FALSE;
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if (LoadTriangleInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],TRUE))
       Goodbye();
     nomeshes = 1;
@@ -1191,8 +1279,10 @@ int main(int argc, char *argv[])
   case 12:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    for(j=0;j<MAXBOUNDARIES;j++) 
-      boundaries[nofile][j].created = FALSE;
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if (LoadMeditInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],TRUE))
       Goodbye();
     nomeshes = 1;
@@ -1201,8 +1291,10 @@ int main(int argc, char *argv[])
   case 13:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    for(j=0;j<MAXBOUNDARIES;j++) 
-      boundaries[nofile][j].created = FALSE;
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if (LoadGidInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],TRUE))
       Goodbye();
     nomeshes = 1;
@@ -1211,21 +1303,31 @@ int main(int argc, char *argv[])
   case 14:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
-    for(j=0;j<MAXBOUNDARIES;j++) 
-      boundaries[nofile][j].created = FALSE;
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
     if (LoadGmshInput(&(data[nofile]),boundaries[nofile],eg.filesin[nofile],TRUE))
       Goodbye();
     nomeshes = 1;
     break;
 
+  case 15: 
+    FuseSolutionElmerPartitioned(eg.filesin[nofile],eg.filesout[nofile],eg.decimals,info);
+    printf("Partitioned solution is not read in memory, thus exiting\n");
+    Goodbye();
+    break;
+
   default:
     Instructions();
     Goodbye();
   }  
 
   nofile++;
-  if(nofile < eg.nofilesin) goto read_another_file;
-
+  if(nofile < eg.nofilesin) {
+    printf("\nElmergrid loading data from another file:\n");
+    goto read_another_file;
+  }
 
   /***********************************/
 
@@ -1234,6 +1336,7 @@ int main(int argc, char *argv[])
 
   printf("\nElmergrid creating and manipulating meshes:\n");
 
+
   if(nogrids > nomeshes && outmethod != 1) { 
 
     nomeshes = nogrids;
@@ -1296,12 +1399,6 @@ int main(int argc, char *argv[])
 			  eg.layerparents, eg.layermove, eg.layereps, info);
   }
 
-  for(k=0;k<nomeshes;k++) {
-    if(eg.polar || data[k].coordsystem == COORD_POLAR) {
-      if(!eg.polar) eg.polarradius = grids[k].polarradius;
-      PolarCoordinates(&data[k],eg.polarradius,info);
-    }
-  }
 
   if(outmethod != 1 && dim != 2 && eg.dim != 2) { 
     j = MAX(nogrids,1);
@@ -1357,6 +1454,18 @@ int main(int argc, char *argv[])
     }
   }
 
+  for(k=0;k<nomeshes;k++) {
+    if(eg.polar || data[k].coordsystem == COORD_POLAR) {
+      if(!eg.polar) eg.polarradius = grids[k].polarradius;
+      PolarCoordinates(&data[k],eg.polarradius,info);
+    }
+  }
+
+  for(k=0;k<nomeshes;k++) {
+    if(eg.cylinder || data[k].coordsystem == COORD_CYL) {
+      CylinderCoordinates(&data[k],info);
+    }
+  }
 
   if(eg.unitemeshes) {
     for(k=1;k<nomeshes;k++)
@@ -1538,49 +1647,72 @@ int main(int argc, char *argv[])
       OptimizePartitioning(&data[k],boundaries[k],info);
   }
 
-  if(eg.pelems) {
+  if(eg.pelems || eg.belems || eg.advancedmat) {
     int currenttype;
-    for(l=0;l<eg.pelems;l++)  
-      printf("Setting element between materials %d and %d to have p=%d.\n",
-	     eg.bulkmap[3*l],eg.bulkmap[3*l+1],eg.bulkmap[3*l+2]);
+
     for(k=0;k<nomeshes;k++) {
       data[k].pelemtypes = Ivector(1,data[k].noelements); 
       data[k].pelems = TRUE;
 
-      for(j=1;j<=data[k].noelements;j++) {
-	currenttype = data[k].material[j];
-	for(l=0;l<eg.pelems;l++) {
-	  if(currenttype >= eg.pelemmap[3*l] && currenttype <= eg.pelemmap[3*l+1]) {
-	    data[k].pelemtypes[j] = eg.pelemmap[3*l+2];
-	    currenttype = -1;
-	  }
-	}
-      }
-    }
-    printf("Creating p-elements finished\n");
-  }
-  if(eg.advancedmat) {
-    int currenttype;
-    for(k=0;k<nomeshes;k++) {
-      data[k].pelemtypes = Ivector(1,data[k].noelements); 
-      data[k].pelems = 2;
       for(j=1;j<=data[k].noelements;j++) 
 	data[k].pelemtypes[j] = 1;
 
-      for(l=0;l<eg.advancedmat;l++) {
-	printf("Setting element of material %d to have advanced settings [%d %d %d %d %d]\n",
-	       eg.advancedelem[6*l],eg.advancedelem[6*l+1],eg.advancedelem[6*l+2],
-	       eg.advancedelem[6*l+3],eg.advancedelem[6*l+4],eg.advancedelem[6*l+5]);
-	
+      if(eg.pelems) {    
+	for(l=0;l<eg.pelems;l++)  
+	  printf("Setting element between materials %d and %d to have p=%d.\n",
+		 eg.pelemmap[3*l],eg.pelemmap[3*l+1],eg.pelemmap[3*l+2]);
+
 	for(j=1;j<=data[k].noelements;j++) {
-	  if( data[k].material[j] == eg.advancedelem[6*l]) {
-	    data[k].pelemtypes[j] = eg.advancedelem[6*l+1] + 10*eg.advancedelem[6*l+2] + 
-	      100*eg.advancedelem[6*l+3] + 1000*eg.advancedelem[6*l+4] + 10000*eg.advancedelem[6*l+5];
+	  currenttype = data[k].material[j];
+	  for(l=0;l<eg.pelems;l++) 
+	    if(currenttype >= eg.pelemmap[3*l] && currenttype <= eg.pelemmap[3*l+1]) {
+	      data[k].pelemtypes[j] += 1000000 * eg.pelemmap[3*l+2];
+	      currenttype = -1;
+	    }
+	}
+	printf("Creating p-elements finished\n");
+      }
+      
+      if(eg.belems) {    
+	for(l=0;l<eg.belems;l++)  
+	  printf("Setting element between materials %d and %d to have bubble dofs=%d.\n",
+		 eg.belemmap[3*l],eg.belemmap[3*l+1],eg.belemmap[3*l+2]);
+
+	for(j=1;j<=data[k].noelements;j++) {
+	  currenttype = data[k].material[j];
+	  for(l=0;l<eg.belems;l++) 
+	    if(currenttype >= eg.belemmap[3*l] && currenttype <= eg.belemmap[3*l+1]) {
+	      data[k].pelemtypes[j] += 10000 * eg.belemmap[3*l+2];
+	      currenttype = -1;
+	    }
+	}
+	printf("Creating bubble elements finished\n");
+      }
+
+      if(eg.advancedmat) {
+	for(l=0;l<eg.advancedmat;l++) {
+	  printf("Setting element of material %d to have advanced settings [%d %d %d %d %d %d]\n",
+		 eg.advancedelem[7*l],eg.advancedelem[7*l+1],eg.advancedelem[7*l+2],
+		 eg.advancedelem[7*l+3],eg.advancedelem[7*l+4],eg.advancedelem[7*l+5],
+		 eg.advancedelem[7*l+6]);
+
+	  for(i=1;i<=6;i++) {
+	    if(eg.advancedelem[7*l+i] < 0 || eg.advancedelem[7*l+i] > 99) {
+	      printf("Advanced elements limited to 9 or 99 (not %d)!\n",eg.advancedelem[7*l+i]);
+	    }
+	  }
+
+	  for(j=1;j<=data[k].noelements;j++) {
+	    if( data[k].material[j] == eg.advancedelem[7*l]) {
+	      data[k].pelemtypes[j] = 1 * eg.advancedelem[7*l+1] + 10*eg.advancedelem[7*l+2] + 
+		100*eg.advancedelem[7*l+3] + 1000*eg.advancedelem[7*l+4] + 
+		10000*eg.advancedelem[7*l+5] + 1000000*eg.advancedelem[7*l+6];
+	    }
 	  }
 	}
+	printf("Creating advanced elements finished\n");
       }
     }
-    printf("Creating advanced elements finished\n");
   }
 
 
