@@ -1,4 +1,29 @@
-/* -------------------------------:  FEMSOLVE.C  :----------------------------
+/*  
+   ElmerGrid - A simple mesh generation and manipulation utility  
+   Copyright (C) 1995- , CSC - Scientific Computing Ltd.   
+
+   Author: Peter Rback
+   Email: Peter.Raback@csc.fi
+   Address: CSC - Scientific Computing Ltd.
+            Keilaranta 14
+            02101 Espoo, Finland
+
+   This program is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License
+   as published by the Free Software Foundation; either version 2
+   of the License, or (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+*/
+
+/* -------------------------------:  femsolve.c  :----------------------------
 
    This module is for solving a dense system of linear equations. For sparse systems
    a special sparse matrix library should be used.
